@@ -16,6 +16,8 @@ from bnlearn.bnlearn import (
     query2df,
     _dag2adjmat,
     _filter_df,
+    save,
+    load,
 )
 
 # Import function in new level
@@ -28,7 +30,7 @@ from packaging import version
 
 __author__ = 'Erdogan Tasksen'
 __email__ = 'erdogant@gmail.com'
-__version__ = '0.3.21'
+__version__ = '0.4.2'
 
 try:
     import pgmpy
