@@ -11,23 +11,21 @@
 import os
 import wget
 import zipfile
-import json
 import itertools
 import pandas as pd
 import numpy as np
 import networkx as nx
 import matplotlib.pyplot as plt
-from ismember import ismember
+from pathlib import Path
 from tqdm import tqdm
 
 from pgmpy.models import BayesianModel
 from pgmpy.factors.discrete import TabularCPD
 from pgmpy.sampling import BayesianModelSampling  # GibbsSampling
-
 from pgmpy import readwrite
-# import bnlearn.network as network
-# from bnlearn.network import compare_networks, graphlayout, adjmat2graph
-# import bnlearn.inference as inference
+
+from ismember import ismember
+import pypickle
 import bnlearn
 
 
@@ -296,6 +294,28 @@ def adjmat2vec(adjmat, min_weight=1):
     return(adjmat)
 
 
+# %%
+def adjmat2dict(adjmat):
+    """Convert adjacency matrix to dict.
+
+    Parameters
+    ----------
+    adjmat : pd.DataFrame
+        Adjacency matrix.
+
+    Returns
+    -------
+    graph : dict
+        Graph.
+
+    """
+    graph={}
+    rows=adjmat.index.values
+    for r in rows:
+        graph.update({r: list(rows[adjmat.loc[r,:]])})
+    return graph
+
+
 # %% Sampling from model
 def sampling(DAG, n=1000, verbose=3):
     """Generate sample(s) using forward sampling from joint distribution of the bayesian network.
@@ -339,6 +359,67 @@ def sampling(DAG, n=1000, verbose=3):
     return(df)
 
 
+# %% Convert BIF model to bayesian model
+def _bif2bayesian(pathname, verbose=3):
+    """Return the fitted bayesian model.
+
+    Example
+    -------
+    >>> from pgmpy.readwrite import BIFReader
+    >>> reader = BIFReader("bif_test.bif")
+    >>> reader.get_model()
+    <pgmpy.models.BayesianModel.BayesianModel object at 0x7f20af154320>
+    """
+    if verbose>=3: print('[bnlearn] >Loading bif file <%s>' %(pathname))
+
+    bifmodel=readwrite.BIF.BIFReader(path=pathname)
+
+    try:
+        model = BayesianModel(bifmodel.variable_edges)
+        model.name = bifmodel.network_name
+        model.add_nodes_from(bifmodel.variable_names)
+
+        tabular_cpds = []
+        for var in sorted(bifmodel.variable_cpds.keys()):
+            values = bifmodel.variable_cpds[var]
+            cpd = TabularCPD(var, len(bifmodel.variable_states[var]), values,
+                             evidence=bifmodel.variable_parents[var],
+                             evidence_card=[len(bifmodel.variable_states[evidence_var])
+                                            for evidence_var in bifmodel.variable_parents[var]])
+            tabular_cpds.append(cpd)
+
+        model.add_cpds(*tabular_cpds)
+#        for node, properties in bifmodel.variable_properties.items():
+#            for prop in properties:
+#                prop_name, prop_value = map(lambda t: t.strip(), prop.split('='))
+#                model.node[node][prop_name] = prop_value
+
+        return model
+
+    except AttributeError:
+        raise AttributeError('[bnlearn] >First get states of variables, edges, parents and network names')
+
+
+# %%
+def query2df(query):
+    """Convert query from inference model to a dataframe.
+
+    Parameters
+    ----------
+    query : Object from the inference model.
+        Convert query object to a dataframe.
+
+    Returns
+    -------
+    df : pd.DataFrame()
+        Dataframe with inferences.
+
+    """
+    df = pd.DataFrame(data = list(itertools.product([0, 1], repeat=len(query.variables))), columns=query.variables)
+    df['p'] = query.values.flatten()
+    return df
+
+
 # %% Model Sprinkler
 def _DAG_sprinkler(CPD=True):
     """Create DAG-model for the sprinkler example.
@@ -383,47 +464,6 @@ def _DAG_sprinkler(CPD=True):
     return(model)
 
 
-# %% Convert BIF model to bayesian model
-def _bif2bayesian(pathname, verbose=3):
-    """Return the fitted bayesian model.
-
-    Example
-    -------
-    >>> from pgmpy.readwrite import BIFReader
-    >>> reader = BIFReader("bif_test.bif")
-    >>> reader.get_model()
-    <pgmpy.models.BayesianModel.BayesianModel object at 0x7f20af154320>
-    """
-    if verbose>=3: print('[bnlearn] >Loading bif file <%s>' %(pathname))
-
-    bifmodel=readwrite.BIF.BIFReader(path=pathname)
-
-    try:
-        model = BayesianModel(bifmodel.variable_edges)
-        model.name = bifmodel.network_name
-        model.add_nodes_from(bifmodel.variable_names)
-
-        tabular_cpds = []
-        for var in sorted(bifmodel.variable_cpds.keys()):
-            values = bifmodel.variable_cpds[var]
-            cpd = TabularCPD(var, len(bifmodel.variable_states[var]), values,
-                             evidence=bifmodel.variable_parents[var],
-                             evidence_card=[len(bifmodel.variable_states[evidence_var])
-                                            for evidence_var in bifmodel.variable_parents[var]])
-            tabular_cpds.append(cpd)
-
-        model.add_cpds(*tabular_cpds)
-#        for node, properties in bifmodel.variable_properties.items():
-#            for prop in properties:
-#                prop_name, prop_value = map(lambda t: t.strip(), prop.split('='))
-#                model.node[node][prop_name] = prop_value
-
-        return model
-
-    except AttributeError:
-        raise AttributeError('[bnlearn] >First get states of variables, edges, parents and network names')
-
-
 # %% Make directed graph from adjmatrix
 def to_undirected(adjmat):
     """Transform directed adjacency matrix to undirected.
@@ -484,7 +524,7 @@ def compare_networks(model_1, model_2, pos=None, showfig=True, figsize=(15, 8),
 
 
 # %% PLOT
-def plot(model, pos=None, scale=1, figsize=(15, 8), verbose=3):
+def plot(model, pos=None, scale=1, figsize=(15, 8), interactive=False, title='bnlearn causal network', params = {'directed':True, 'height':'800px', 'width':'70%', 'notebook':False, 'layout':None, 'font_color': False, 'bgcolor':'#ffffff'}, verbose=3):
     """Plot the learned stucture.
 
     Parameters
@@ -497,6 +537,11 @@ def plot(model, pos=None, scale=1, figsize=(15, 8), verbose=3):
         Scaling parameter for the network. A larger number will linearily increase the network.. The default is 1.
     figsize : tuple, optional
         Figure size. The default is (15,8).
+    interactive : Bool, (default: True)
+        True: Static plot
+        False: Interactive web-based graph.
+    params : dict
+        Dictionary containing parameters for the pyvis network. This only works inf interactive=True
     verbose : int, optional
         Print progress to screen. The default is 3.
         0: None, 1: ERROR, 2: WARN, 3: INFO (default), 4: DEBUG, 5: TRACE
@@ -510,12 +555,16 @@ def plot(model, pos=None, scale=1, figsize=(15, 8), verbose=3):
             Graph model
 
     """
+    defaults = {'directed':True, 'height':'800px', 'width':'70%', 'notebook':False, 'heading':title, 'layout':None, 'font_color': False, 'bgcolor':'#ffffff'}
+    params = {**defaults, **params}
+
     out = {}
     G = nx.DiGraph()  # Directed graph
     layout='fruchterman_reingold'
 
     # Extract model if in dict
     if 'dict' in str(type(model)):
+        adjmat = model.get('adjmat', None)
         model = model.get('model', None)
 
     # Bayesian model
@@ -531,37 +580,59 @@ def plot(model, pos=None, scale=1, figsize=(15, 8), verbose=3):
     elif 'networkx' in str(type(model)):
         if verbose>=3: print('[bnlearn] >Plot based on networkx model')
         G = model
-        pos = graphlayout(G, pos=pos, scale=scale, layout=layout, verbose=verbose)
+        pos = bnlearn.network.graphlayout(G, pos=pos, scale=scale, layout=layout, verbose=verbose)
     else:
         if verbose>=3: print('[bnlearn] >Plot based on adjacency matrix')
-        G = adjmat2graph(model)
+        G = bnlearn.network.adjmat2graph(adjmat)
         # Get positions
-        pos = graphlayout(G, pos=pos, scale=scale, layout=layout, verbose=verbose)
+        pos = bnlearn.network.graphlayout(G, pos=pos, scale=scale, layout=layout, verbose=verbose)
 
-    # Bootup figure
-    plt.figure(figsize=figsize)
-    # nodes
-    nx.draw_networkx_nodes(G, pos, node_size=500, alpha=0.85)
-    # edges
-    colors = [G[u][v].get('color', 'k') for u, v in G.edges()]
-    weights = [G[u][v].get('weight', 1) for u, v in G.edges()]
-    nx.draw_networkx_edges(G, pos, arrowstyle='->', edge_color=colors, width=weights)
-    # Labels
-    nx.draw_networkx_labels(G, pos, font_size=20, font_family='sans-serif')
-    # Get labels of weights
-    # labels = nx.get_edge_attributes(G,'weight')
-    # Plot weights
-    nx.draw_networkx_edge_labels(G, pos, edge_labels=nx.get_edge_attributes(G, 'weight'))
-    # Making figure nice
-    ax = plt.gca()
-    ax.set_axis_off()
-    plt.show()
+    # Make interactive or static plot
+    if interactive:
+        try:
+            from pyvis import network as net
+            from IPython.core.display import display, HTML
+            # Convert adjacency matrix into Networkx Graph
+            G = bnlearn.network.adjmat2graph(adjmat)
+            # Setup of the interactive network figure
+            g = net.Network(**params)
+            # g = net.Network(directed=True, height='800px', width='70%', notebook=False, heading=title)
+            g.from_nx(G)
+            # Create advanced buttons
+            g.show_buttons(filter_=['physics'])
+            # Display
+            filename = title.strip().replace(' ','_') + '.html'
+            g.show(filename)
+            display(HTML(filename))
+            # webbrowser.open('bnlearn.html')
+        except ModuleNotFoundError:
+            if verbose>=2: print('[bnlearn] >"pyvis" module is not installed. Please pip install first: "pip install pyvis"')
+    else:
+        # Bootup figure
+        plt.figure(figsize=figsize)
+        # nodes
+        nx.draw_networkx_nodes(G, pos, node_size=500, alpha=0.85)
+        # edges
+        colors = [G[u][v].get('color', 'k') for u, v in G.edges()]
+        weights = [G[u][v].get('weight', 1) for u, v in G.edges()]
+        nx.draw_networkx_edges(G, pos, arrowstyle='->', edge_color=colors, width=weights)
+        # Labels
+        nx.draw_networkx_labels(G, pos, font_size=20, font_family='sans-serif')
+        # Get labels of weights
+        # labels = nx.get_edge_attributes(G,'weight')
+        # Plot weights
+        nx.draw_networkx_edge_labels(G, pos, edge_labels=nx.get_edge_attributes(G, 'weight'))
+        # Making figure nice
+        ax = plt.gca()
+        ax.set_axis_off()
+        plt.show()
 
     # Store
     out['pos']=pos
     out['G']=G
     return(out)
 
+
 # %%
 def topological_sort(adjmat, start=None):
     """Topological sort.
@@ -624,28 +695,6 @@ def topological_sort(adjmat, start=None):
     return stack + order[::-1]
 
 
-# %%
-def adjmat2dict(adjmat):
-    """Convert adjacency matrix to dict.
-
-    Parameters
-    ----------
-    adjmat : pd.DataFrame
-        Adjacency matrix.
-
-    Returns
-    -------
-    graph : dict
-        Graph.
-
-    """
-    graph={}
-    rows=adjmat.index.values
-    for r in rows:
-        graph.update({r: list(rows[adjmat.loc[r,:]])})
-    return graph
-
-
 # %% Example data
 def import_example(data='sprinkler', n=10000, verbose=3):
     """Load example dataset.
@@ -688,6 +737,7 @@ def import_example(data='sprinkler', n=10000, verbose=3):
 
     return df
 
+
 #%% Download data from github source
 def _download_example(data, verbose=3):
     # Set url location
@@ -706,6 +756,7 @@ def _download_example(data, verbose=3):
     
     return PATH_TO_DATA
 
+
 # %% Make DAG
 def import_DAG(filepath='sprinkler', CPD=True, checkmodel=True, verbose=3):
     """Import Directed Acyclic Graph.
@@ -963,84 +1014,71 @@ def _get_prob(query, method='max'):
     return out
 
 
-# %%
-def query2df(query):
-    """Convert query from inference model to a dataframe.
+# %% Save model
+def save(model, filepath='bnlearn_model.pkl', overwrite=False, verbose=3):
+    """Save learned model in pickle file.
 
     Parameters
     ----------
-    query : Object from the inference model.
-        Convert query object to a dataframe.
+    filepath : str, (default: 'bnlearn_model.pkl')
+        Pathname to store pickle files.
+    overwrite : bool, (default=False)
+        Overwite file if exists.
+    verbose : int, optional
+        Show message. A higher number gives more informatie. The default is 3.
 
     Returns
     -------
-    df : pd.DataFrame()
-        Dataframe with inferences.
+    bool : [True, False]
+        Status whether the file is saved.
 
     """
-    df = pd.DataFrame(data = list(itertools.product([0, 1], repeat=len(query.variables))), columns=query.variables)
-    df['p'] = query.values.flatten()
-    return df
+    if (filepath is None) or (filepath==''):
+        filepath = 'bnlearn_model.pkl'
+    if filepath[-4:] != '.pkl':
+        filepath = filepath + '.pkl'
+    filepath = str(Path(filepath).absolute())
 
-# %% Comparison of two networks
-# def _compare_networks(adjmat_true, adjmat_pred, pos=None, showfig=True, width=15, height=8, verbose=3):
-#     # Make sure columns and indices to match
-#     [IArow,IBrow]=ismember(adjmat_true.index.values, adjmat_pred.index.values)
-#     [IAcol,IBcol]=ismember(adjmat_true.columns.values, adjmat_pred.columns.values)
-#     adjmat_true = adjmat_true.loc[IArow,IAcol]
-#     adjmat_pred = adjmat_pred.iloc[IBrow,IBcol]
-    
-#     # Make sure it is boolean adjmat
-#     adjmat_true = adjmat_true>0
-#     adjmat_pred = adjmat_pred>0
+    # Store data
+    # storedata = {}
+    # storedata['model'] = model
+    # Save
+    status = pypickle.save(filepath, model, overwrite=overwrite, verbose=verbose)
+    # return
+    return status
 
-#     # Check whether order is correct
-#     assert np.all(adjmat_true.columns.values==adjmat_pred.columns.values), 'Column order of both input values could not be matched'
-#     assert np.all(adjmat_true.index.values==adjmat_pred.index.values), 'Row order of both input values could not be matched'
-    
-#     # Get edges
-#     y_true = adjmat_true.stack().reset_index()[0].values
-#     y_pred = adjmat_pred.stack().reset_index()[0].values
-    
-#     # Confusion matrix
-#     scores=confmatrix.twoclass(y_true, y_pred, threshold=0.5, classnames=['Disconnected','Connected'], title='', cmap=plt.cm.Blues, showfig=1, verbose=0)
-#     #bayes.plot(out_bayes['adjmat'], pos=G['pos'])
-    
-#     # Setup graph
-#     adjmat_diff = adjmat_true.astype(int)
-#     adjmat_diff[(adjmat_true.astype(int) - adjmat_pred.astype(int))<0]=2
-#     adjmat_diff[(adjmat_true.astype(int) - adjmat_pred.astype(int))>0]=-1
-    
-#     if showfig:
-#         # Setup graph
-#     #    G_true = adjmat2graph(adjmat_true)
-#         G_diff = adjmat2graph(adjmat_diff)
-#         # Graph layout
-#         pos = graphlayout(G_diff, pos=pos, scale=1, layout='fruchterman_reingold', verbose=verbose)
-#         # Bootup figure
-#         plt.figure(figsize=(width,height))
-#         # nodes
-#         nx.draw_networkx_nodes(G_diff, pos, node_size=700)
-#         # edges
-#         colors  = [G_diff[u][v]['color'] for u,v in G_diff.edges()]
-#         #weights = [G_diff[u][v]['weight'] for u,v in G_diff.edges()]
-#         nx.draw_networkx_edges(G_diff, pos, arrowstyle='->', edge_color=colors, width=1)
-#         # Labels
-#         nx.draw_networkx_labels(G_diff, pos, font_size=20, font_family='sans-serif')
-#         # Get labels of weights
-#         #labels = nx.get_edge_attributes(G,'weight')
-#         # Plot weights
-#         nx.draw_networkx_edge_labels(G_diff, pos, edge_labels=nx.get_edge_attributes(G_diff,'weight'))
-#         # Making figure nice
-#         #plt.legend(['Nodes','TN','FP','test'])
-#         ax = plt.gca()
-#         ax.set_axis_off()
-#         plt.show()
-    
-#     # Return
-#     return(scores, adjmat_diff)
 
-# # Make graph layout
+# %% Load model.
+def load(filepath='bnlearn_model.pkl', verbose=3):
+    """Load learned model.
+
+    Parameters
+    ----------
+    filepath : str
+        Pathname to stored pickle files.
+    verbose : int, optional
+        Show message. A higher number gives more information. The default is 3.
+
+    Returns
+    -------
+    Object.
+
+    """
+    if (filepath is None) or (filepath==''):
+        filepath = 'bnlearn_model.pkl'
+    if filepath[-4:]!='.pkl':
+        filepath = filepath + '.pkl'
+    filepath = str(Path(filepath).absolute())
+    # Load
+    model = pypickle.load(filepath, verbose=verbose)
+    # Store in self.
+    if model is not None:
+        return model
+    else:
+        if verbose>=2: print('[bnlearn] >WARNING: Could not load data from [%s]' %(filepath))
+
+
+# %% Make graph layout
 # def graphlayout(model, pos, scale=1, layout='fruchterman_reingold', verbose=3):
 #     if isinstance(pos, type(None)):
 #         if layout=='fruchterman_reingold':
@@ -1051,28 +1089,3 @@ def query2df(query):
 #         if verbose>=3: print('[bnlearn] >Existing coordinates from <pos> are used.')
 
 #     return(pos)
-
-# def adjmat2graph(adjmat):
-#     G = nx.DiGraph() # Directed graph
-#     # Convert adjmat to source target
-#     df_edges=adjmat.stack().reset_index()
-#     df_edges.columns=['source', 'target', 'weight']
-#     df_edges['weight']=df_edges['weight'].astype(float)
-    
-#     # Add directed edge with weigth
-#     for i in range(df_edges.shape[0]):
-#         if df_edges['weight'].iloc[i]!=0:
-#             # Setup color
-#             if df_edges['weight'].iloc[i]==1:
-#                 color='k'
-#             elif df_edges['weight'].iloc[i]>1:
-#                 color='r'
-#             elif df_edges['weight'].iloc[i]<0:
-#                 color='b'
-#             else:
-#                 color='p'
-            
-#             # Create edge in graph
-#             G.add_edge(df_edges['source'].iloc[i], df_edges['target'].iloc[i], weight=np.abs(df_edges['weight'].iloc[i]), color=color)    
-#     # Return
-#     return(G)
