@@ -1,19 +1,169 @@
 # %%
 import bnlearn as bn
-# print(bn.__version__)
+print(bn.__version__)
 print(dir(bn))
 
-# print(dir(bn.structure_learning))
-# print(dir(bn.parameter_learning))
-# print(dir(bn.inference))
+print(dir(bn.structure_learning))
+print(dir(bn.parameter_learning))
+print(dir(bn.inference))
+
+# %% Save and load trained models
+# Import example
+df = bn.import_example(data='asia')
+# Learn structure
+model = bn.structure_learning.fit(df, methodtype='tan', class_node='lung')
+# Save model
+bn.save(model, filepath='bnlearn_model', overwrite=True)
+# Load model
+model = bn.load(filepath='bnlearn_model')
 
 
-# %%
+# %% CHECK DIFFERENCES PGMPY vs. BNLEARN
+import pandas as pd
+import numpy as np
+import matplotlib.pyplot as plt
+from pgmpy.estimators import TreeSearch
+from pgmpy.models import BayesianModel
+import networkx as nx
+import bnlearn as bnlearn
+from pgmpy.inference import VariableElimination
+from pgmpy.estimators import BDeuScore, K2Score, BicScore
 import bnlearn as bn
-examples = ['titanic', 'sprinkler', 'alarm', 'andes', 'asia', 'sachs', 'water', 'miserables']
-for example in examples:
-    df = bn.import_DAG(example)
-    # assert ~df.empty
+
+
+df=bnlearn.import_example(data='andes')
+
+# PGMPY
+est = TreeSearch(df)
+dag = est.estimate(estimator_type="tan",class_node='DISPLACEM0')
+bnq = BayesianModel(dag.edges())
+bnq.fit(df, estimator=None) # None means maximum likelihood estimator
+bn_infer = VariableElimination(bnq)
+q = bn_infer.query(variables=['DISPLACEM0'], evidence={'RApp1':1})
+print(q)
+
+# BNLEARN
+model = bnlearn.structure_learning.fit(df, methodtype='tan', class_node='DISPLACEM0', scoretype='bic')
+model_bn = bnlearn.parameter_learning.fit(model, df, methodtype='ml') # maximum likelihood estimator
+query=bnlearn.inference.fit(model_bn, variables=['DISPLACEM0'], evidence={'RApp1':1})
+
+# DAG COMPARISON
+assert np.all(model_bn['adjmat']==model['adjmat'])
+assert dag.edges()==model['model'].edges()
+assert dag.edges()==model['model_edges']
+
+# COMPARE THE CPDs names
+qbn_cpd = []
+bn_cpd = []
+for cpd in bnq.get_cpds(): qbn_cpd.append(cpd.variable)
+for cpd in model_bn['model'].get_cpds(): bn_cpd.append(cpd.variable)
+
+assert len(bn_cpd)==len(qbn_cpd)
+assert np.all(np.isin(bn_cpd, qbn_cpd))
+
+# COMPARE THE CPD VALUES
+nr_diff = 0
+for cpd_bnlearn in model_bn['model'].get_cpds():
+    for cpd_pgmpy in bnq.get_cpds():
+        if cpd_bnlearn.variable==cpd_pgmpy.variable:
+            assert np.all(cpd_bnlearn.values==cpd_pgmpy.values)
+            # if not np.all(cpd_bnlearn.values==cpd_pgmpy.values):
+                # print('%s-%s'%(cpd_bnlearn.variable, cpd_pgmpy.variable))
+                # print(cpd_bnlearn)
+                # print(cpd_pgmpy)
+                # nr_diff=nr_diff+1
+                # input('press enter to see the next difference in CPD.')
+
+
+#%% Example of interactive plotting
+from pgmpy.estimators import TreeSearch
+from pgmpy.models import BayesianModel
+import bnlearn as bn
+
+# Import example
+df = bn.import_example(data='asia')
+
+# Do the tan learning
+est = TreeSearch(df)
+dag = est.estimate(estimator_type="tan",class_node='lung')
+
+# And now with bnlearn
+model = bn.structure_learning.fit(df, methodtype='tan', class_node='lung')
+
+# Compare results
+assert dag.edges()==model['model'].edges()
+assert dag.edges()==model['model_edges']
+
+
+
+from pgmpy.inference import VariableElimination
+bnp = BayesianModel(dag.edges())
+bn_infer = VariableElimination(bnp)
+bnp.fit(df)
+
+a = bn_infer.query(variables=['Target'], evidence={'X':'c'})
+print(a)
+
+#%% Example of interactive plotting
+import bnlearn as bn
+
+# Load example dataset
+df = bn.import_example(data='asia')
+
+# Structure learning
+model = bn.structure_learning.fit(df)
+
+bn.plot(model, interactive=False)
+
+# Add some parameters for the interactive plot
+bn.plot(model, interactive=True, params = {'height':'600px'})
+
+# Add more parameters for the interactive plot
+bn.plot(model, interactive=True, params = {'directed':True, 'height':'800px', 'width':'70%', 'notebook':False, 'heading':'bnlearn causal diagram', 'layout':None, 'font_color': False, 'bgcolor':'#ffffff'})
+
+# %% TAN : Tree-augmented Naive Bayes (TAN)
+# https://pgmpy.org/examples/Structure%20Learning%20with%20TAN.html
+import bnlearn as bn
+
+df = bn.import_example()
+# Structure learning
+model = bn.structure_learning.fit(df)
+# bn.plot(model)
+bn.plot(model, interactive=True)
+# bn.plot(model, interactive=True, params = {'height':'800px'})
+
+
+# %% Large dataset
+import pandas as pd
+df=pd.read_csv('c:/temp/features.csv')
+model = bn.structure_learning.fit(df.iloc[:,0:1000], methodtype='cl', root_node='21')
+model = bn.structure_learning.fit(df.iloc[:,0:100], methodtype='cs')
+bn.plot(model)
+
+# %% White_list edges
+import bnlearn as bn
+DAG = bn.import_DAG('asia')
+# plot ground truth
+df = bn.sampling(DAG, n=1000)
+
+# Structure learning with black list
+model = bn.structure_learning.fit(df, methodtype='hc', white_list=[('tub','lung'), ('smoke','bronc')], bw_list_method='edges')
+bn.plot(model)
+model = bn.structure_learning.fit(df, methodtype='hc', white_list=['tub','lung', 'smoke','bronc'], bw_list_method='nodes')
+bn.plot(model)
+model = bn.structure_learning.fit(df, methodtype='hc')
+bn.plot(model)
+
+
+# %% TAN : Tree-augmented Naive Bayes (TAN)
+# https://pgmpy.org/examples/Structure%20Learning%20with%20TAN.html
+import bnlearn as bn
+
+df = bn.import_example()
+# Structure learning
+model = bn.structure_learning.fit(df, methodtype='tan', root_node='Cloudy', class_node='Rain', verbose=0)
+bn.plot(model)
+bn.plot(model, interactive=True)
 
 # %% Download example
 import bnlearn as bn
@@ -80,7 +230,7 @@ bn.topological_sort(bn.adjmat2vec(DAG['adjmat']), 'Rain')
 # %%
 
 import bnlearn as bn
-DAG = bn.import_DAG('sprinkler', verbose=0)
+DAG = bn.import_DAG('sprinkler')
 df = bn.sampling(DAG, n=1000, verbose=0)
 model = bn.structure_learning.fit(df, methodtype='chow-liu', root_node='Wet_Grass')
 G = bn.plot(model)
@@ -95,6 +245,7 @@ model = bn.structure_learning.fit(df)
 G = bn.plot(model)
 
 # %% Load example dataframe from sprinkler
+import bnlearn as bn
 DAG = bn.import_DAG('sprinkler', verbose=0)
 df = bn.sampling(DAG, n=1000, verbose=0)
 
@@ -120,7 +271,8 @@ to_vector = bn.adjmat2vec(DAG['adjmat'])
 to_adjmat = bn.vec2adjmat(to_vector['source'], to_vector['target'])
 
 # %% Load example dataframe from sprinkler
-df = bn.import_example('sprinkler', verbose=0)
+import bnlearn as bn
+df = bn.import_example('sprinkler')
 # Structure learning
 model = bn.structure_learning.fit(df, verbose=0)
 # Plot
@@ -173,8 +325,11 @@ edges = [('smoke', 'lung'),
 
 # Make the actual Bayesian DAG
 DAG = bn.make_DAG(edges, verbose=0)
+bn.save(DAG, overwrite=True)
+DAG1 = bn.load()
+
 # Plot the DAG
-bn.plot(DAG, verbose=0)
+bn.plot(DAG1, verbose=0)
 # Print the CPDs
 bn.print_CPD(DAG)
 
@@ -261,19 +416,22 @@ DAG = bn.import_DAG('asia')
 G = bn.plot(DAG)
 # Sampling
 df = bn.sampling(DAG, n=10000)
+
+# Structure learning
+model = bn.structure_learning.fit(df, verbose=0)
 # Structure learning of sampled dataset
 model_sl = bn.structure_learning.fit(df, methodtype='hc', scoretype='bic')
 # Plot based on structure learning of sampled data
-bn.plot(model_sl, pos=G['pos'])
+bn.plot(model_sl, pos=G['pos'], interactive=True, params = {'height':'800px'})
 # Compare networks and make plot
 bn.compare_networks(model, model_sl, pos=G['pos'])
 
 
 # Structure learning with black list
-model_wl = bn.structure_learning.fit(df, methodtype='hc', white_list=['asia','tub','bronc','xray','smoke'], bw_list_method='enforce')
+model_wl = bn.structure_learning.fit(df, methodtype='hc', white_list=['asia','tub','bronc','xray','smoke'], bw_list_method='edges')
 bn.plot(model_wl, pos=G['pos'])
 
-model_bl = bn.structure_learning.fit(df, methodtype='hc', black_list=['asia','tub'], bw_list_method='enforce')
+model_bl = bn.structure_learning.fit(df, methodtype='hc', black_list=['asia','tub'], bw_list_method='edges')
 bn.plot(model_bl, pos=G['pos'])
 
 # Compare models
@@ -353,8 +511,8 @@ df_raw = bn.import_example(data='titanic')
 # Convert to onehot
 dfhot, dfnum = bn.df2onehot(df_raw)
 # Structure learning
-# DAG = bn.structure_learning.fit(dfnum, methodtype='cl', black_list=['Embarked','Parch','Name'], root_node='Survived', bw_list_method='filter')
-DAG = bn.structure_learning.fit(dfnum, methodtype='hc', black_list=['Embarked','Parch','Name'], bw_list_method='filter')
+# DAG = bn.structure_learning.fit(dfnum, methodtype='cl', black_list=['Embarked','Parch','Name'], root_node='Survived', bw_list_method='nodes')
+DAG = bn.structure_learning.fit(dfnum, methodtype='hc', black_list=['Embarked','Parch','Name'], bw_list_method='edges')
 # Plot
 G = bn.plot(DAG)
 # Parameter learning
