@@ -21,8 +21,9 @@ Currently, the library supports parameter learning for *discrete* nodes:
 
 # %% Libraries
 from pgmpy.estimators import MaximumLikelihoodEstimator, BayesianEstimator
-from bnlearn.bnlearn import _filter_df
-from bnlearn.bnlearn import to_BayesianModel
+import bnlearn
+# from bnlearn.bnlearn import _filter_df
+# from bnlearn.bnlearn import to_bayesianmodel
 import warnings
 warnings.filterwarnings("ignore")
 
@@ -111,16 +112,16 @@ def fit(model, df, methodtype='bayes', verbose=3):
 
     # Check whether all labels in the adjacency matrix are included from the dataframe
     # adjmat, model = _check_adjmat(model, df)
-    df = _filter_df(adjmat, df, verbose=config['verbose'])
+    df = bnlearn._filter_df(adjmat, df, verbose=config['verbose'])
 
-    if config['verbose']>=3: print('[BNLEARN][PARAMETER LEARNING] Computing parameters using [%s]' %(config['method']))
+    if config['verbose']>=3: print('[bnlearn] >Parameter learning> Computing parameters using [%s]' %(config['method']))
     # Extract model
     if isinstance(model, dict):
         model = model['model']
 
     # Convert to BayesianModel
     if 'BayesianModel' not in str(type(model)):
-        model = to_BayesianModel(adjmat, verbose=config['verbose'])
+        model = bnlearn.to_bayesianmodel(adjmat, verbose=config['verbose'])
 
     # pe = ParameterEstimator(model, df)
     # print("\n", pe.state_counts('Cloudy'))
@@ -128,17 +129,19 @@ def fit(model, df, methodtype='bayes', verbose=3):
 
     # Learning CPDs using Maximum Likelihood Estimators
     if config['method']=='ml' or config['method']=='maximumlikelihood':
-        # mle = MaximumLikelihoodEstimator(model, df)
-        model = MaximumLikelihoodEstimator(model, df)
-        for node in model.state_names:
-            print(model.estimate_cpd(node))
+        # model = MaximumLikelihoodEstimator(model, df)
+        # for node in model.state_names:
+        #     print(model.estimate_cpd(node))
+        model.fit(df, estimator=None) # estimator as None makes it maximum likelihood estimator according pgmpy docs. 
+        for cpd in model.get_cpds():
+              if config['verbose']>=3: print("[bnlearn] >CPD of {variable}:".format(variable=cpd.variable))
+              if config['verbose']>=3: print(cpd)
 
     #  Learning CPDs using Bayesian Parameter Estimation
     if config['method']=='bayes':
         model.fit(df, estimator=BayesianEstimator, prior_type="BDeu", equivalent_sample_size=1000)
-
         for cpd in model.get_cpds():
-            if config['verbose']>=3: print("CPD of {variable}:".format(variable=cpd.variable))
+            if config['verbose']>=3: print("[bnlearn] >CPD of {variable}:".format(variable=cpd.variable))
             if config['verbose']>=3: print(cpd)
 
     out = {}
