@@ -13,7 +13,7 @@ import numpy as np
 import matplotlib.pyplot as plt
 
 from pgmpy.estimators import BDeuScore, K2Score, BicScore
-from pgmpy.estimators import ExhaustiveSearch, HillClimbSearch
+from pgmpy.estimators import ExhaustiveSearch, HillClimbSearch, TreeSearch
 
 import pgmpy
 from packaging import version
@@ -28,7 +28,7 @@ import bnlearn
 
 
 # %% Structure Learning
-def fit(df, methodtype='hc', scoretype='bic', black_list=None, white_list=None, bw_list_method=None, max_indegree=None, tabu_length=100, epsilon=1e-4, max_iter=1e6, root_node=None, verbose=3):
+def fit(df, methodtype='hc', scoretype='bic', black_list=None, white_list=None, bw_list_method=None, max_indegree=None, tabu_length=100, epsilon=1e-4, max_iter=1e6, root_node=None, class_node=None, fixed_edges=None, verbose=3):
     """Structure learning fit model.
 
     Description
@@ -41,6 +41,7 @@ def fit(df, methodtype='hc', scoretype='bic', black_list=None, white_list=None,
             a. exhaustivesearch
             b. hillclimbsearch
             c. chow-liu
+            d. Tree-augmented Naive Bayes (tan)
         2. Constraint-based structure learning (PC)
             a. chi-square test
         3. Hybrid structure learning (The combination of both techniques) (MMHC)
@@ -49,9 +50,8 @@ def fit(df, methodtype='hc', scoretype='bic', black_list=None, white_list=None,
     This approach construes model selection as an optimization task. It has two building blocks:
     A scoring function sD:->R that maps models to a numerical score, based on how well they fit to a given data set D.
     A search strategy to traverse the search space of possible models M and select a model with optimal score.
-    Commonly used scoring functions to measure the fit between model and data are Bayesian Dirichlet scores such as BDeu or K2
-    and the Bayesian Information Criterion (BIC, also called MDL).
-    As before, BDeu is dependent on an equivalent sample size.
+    Commonly used scoring functions to measure the fit between model and data are Bayesian Dirichlet scores such as BDeu or K2 and the Bayesian Information Criterion (BIC, also called MDL).
+    BDeu is dependent on an equivalent sample size.
 
     Parameters
     ----------
@@ -63,30 +63,36 @@ def fit(df, methodtype='hc', scoretype='bic', black_list=None, white_list=None,
         'ex' or 'exhaustivesearch'
         'cs' or 'constraintsearch'
         'cl' or 'chow-liu' (requires setting root_node parameter)
+        'tan' (Tree-augmented Naive Bayes, requires setting root_node and class_node parameter)
     scoretype : str, (default : 'bic')
         Scoring function for the search spaces.
         'bic', 'k2', 'bdeu'
     black_list : List or None, (default : None)
-        List of edges are black listed by either by filtering or enforcing (see bw_list_method).
-        In case of filtering, the nodes black listed nodes are removed from the dataframe. The resulting model will nog contain any edges that are in black_list.
+        List of edges are black listed.
+        In case of filtering on nodes, the nodes black listed nodes are removed from the dataframe. The resulting model will not contain any nodes that are in black_list.
     white_list : List or None, (default : None)
-        List of edges are white listed by either by filtering or enforcing (see bw_list_method).
-        In case of filtering, the search is limited to those edges. The resulting model will then only contain edges that are in white_list.
+        List of edges are white listed.
+        In case of filtering on nodes, the search is limited to those edges. The resulting model will then only contain nodes that are in white_list.
         Works only in case of methodtype='hc' See also paramter: `bw_list_method`
-    bw_list_method : str, (default : None)
-        'enforce' : A list of edges can optionally be passed as `black_list` or `white_list` to exclude those edges or to limit the search. This option is limited to only methodtype='hc'
-        'filter' : Filter the dataframe based on `black_list` or `white_list`. Filtering can be done for every methodtype/scoretype.
+    bw_list_method : list of str or tuple, (default : None)
+        A list of edges can be passed as `black_list` or `white_list` to exclude or to limit the search. 
+            * 'edges' : [('A', 'B'), ('C','D'), (...)] This option is limited to only methodtype='hc'
+            * 'nodes' : ['A', 'B', ...] Filter the dataframe based on the nodes for `black_list` or `white_list`. Filtering can be done for every methodtype/scoretype.
     max_indegree : int, (default : None)
         If provided and unequal None, the procedure only searches among models where all nodes have at most max_indegree parents. (only in case of methodtype='hc')
     epsilon: float (default: 1e-4)
         Defines the exit condition. If the improvement in score is less than `epsilon`, the learned model is returned. (only in case of methodtype='hc')
     max_iter: int (default: 1e6)
         The maximum number of iterations allowed. Returns the learned model when the number of iterations is greater than `max_iter`. (only in case of methodtype='hc')
-    root_node: String
-        The root node for chow-liu method.
+    root_node: String. (only in case of chow-liu, Tree-augmented Naive Bayes (TAN))
+        The root node for treeSearch based methods.
+    class_node: String
+        The class node is required for Tree-augmented Naive Bayes (TAN)
+    fixed_edges: iterable, Only in case of HillClimbSearch.
+        A list of edges that will always be there in the final learned model. The algorithm will add these edges at the start of the algorithm and will never change it.
     verbose : int, (default : 3)
         Print progress to screen.
-        0: NONE, 1: ERROR,  2: WARNING, 3: INFO (default), 4: DEBUG, 5: TRACE
+        0: None, 1: Error,  2: Warning, 3: Info (default), 4: Debug, 5: Trace
 
     Returns
     -------
@@ -117,12 +123,24 @@ def fit(df, methodtype='hc', scoretype='bic', black_list=None, white_list=None,
     """
     assert isinstance(pd.DataFrame(), type(df)), 'df must be of type pd.DataFrame()'
     assert (scoretype=='bic') | (scoretype=='k2') | (scoretype=='bdeu'), 'scoretype must be string: "bic", "k2" or "bdeu"'
-    assert (methodtype=='cl') | (methodtype=='chow-liu') | (methodtype=='hc') | (methodtype=='ex') | (methodtype=='cs') | (methodtype=='exhaustivesearch')| (methodtype=='hillclimbsearch')| (methodtype=='constraintsearch'), 'Methodtype string is invalid'  # noqa
+    assert (methodtype=='tan') | (methodtype=='cl') | (methodtype=='chow-liu') | (methodtype=='hc') | (methodtype=='ex') | (methodtype=='cs') | (methodtype=='exhaustivesearch')| (methodtype=='hillclimbsearch')| (methodtype=='constraintsearch'), 'Methodtype string is invalid'  # noqa
     if isinstance(white_list, str): white_list = [white_list]
     if isinstance(black_list, str): black_list = [black_list]
     if (white_list is not None) and len(white_list)==0: white_list = None
     if (black_list is not None) and len(black_list)==0: black_list = None
-    if (methodtype!='hc') and (bw_list_method=='enforce'): raise Exception('[bnlearn] >The bw_list_method="%s" does not work with methodtype="%s"' %(bw_list_method, methodtype))
+    if (methodtype!='hc') and (bw_list_method=='edges'): raise Exception('[bnlearn] >The bw_list_method="%s" does not work with methodtype="%s"' %(bw_list_method, methodtype))
+    if (methodtype=='tan') and (class_node is None): raise Exception('[bnlearn] >The treeSearch method TAN requires setting the <class_node> parameter: "%s"' %(str(class_node)))
+    if methodtype=='cl': methodtype = 'chow-liu'
+    if fixed_edges is None: fixed_edges=set()
+
+    # Remove in future version
+    if bw_list_method=='filter':
+        if verbose>=2: print('[bnlearn]> Warning: The parameter bw_list_method="filter" is changed into bw_list_method="nodes", but will be removed in future releases.')
+        bw_list_method = "nodes"
+    if bw_list_method=='enforce':
+        if verbose>=2: print('[bnlearn]> Warning: The parameter bw_list_method="enforce" is changed into bw_list_method="edges", but will be removed in future releases.')
+        bw_list_method = "edges"
+    ## End removal
 
     config = {}
     config['verbose'] = verbose
@@ -136,6 +154,8 @@ def fit(df, methodtype='hc', scoretype='bic', black_list=None, white_list=None,
     config['epsilon'] = epsilon
     config['max_iter'] = max_iter
     config['root_node'] = root_node
+    config['class_node'] = class_node
+    config['fixed_edges'] = fixed_edges
 
     # Show warnings
     # PGMPY_VER = version.parse(pgmpy.__version__)>version.parse("0.1.9")  # Can be be removed if pgmpy >v0.1.9
@@ -151,6 +171,8 @@ def fit(df, methodtype='hc', scoretype='bic', black_list=None, white_list=None,
     #     if config['verbose']>=2: print('[bnlearn] >Warning: white_list only works in case of methodtype="hc"')
     if (max_indegree is not None) and methodtype!='hc':
         if config['verbose']>=2: print('[bnlearn] >Warning: max_indegree only works in case of methodtype="hc"')
+    if (class_node is not None) and methodtype!='tan':
+        if config['verbose']>=2: print('[bnlearn] >Warning: max_indegree only works in case of methodtype="tan"')
 
     if config['verbose']>=3: print('[bnlearn] >Computing best DAG using [%s]' %(config['method']))
 
@@ -181,6 +203,7 @@ def fit(df, methodtype='hc', scoretype='bic', black_list=None, white_list=None,
                                bw_list_method=bw_list_method,
                                epsilon=config['epsilon'],
                                max_iter=config['max_iter'],
+                               fixed_edges=config['fixed_edges'],
                                verbose=config['verbose'],
                                )
 
@@ -193,10 +216,11 @@ def fit(df, methodtype='hc', scoretype='bic', black_list=None, white_list=None,
         Independencies in the data can be identified using chi2 conditional independence tests."""
         out = _constraintsearch(df, verbose=config['verbose'])
 
-    # Constraint-based Structure Learning
-    if config['method']=='cl' or config['method']=='chow-liu':
-        """Chow-liu based Structure Learning."""
-        out = _chowliu(df, config['root_node'], verbose=config['verbose'])
+    # TreeSearch-based Structure Learning
+    if config['method']=='chow-liu' or config['method']=='tan':
+        """TreeSearch based Structure Learning."""
+        out = _treesearch(df, config['method'], config['root_node'], class_node=config['class_node'], verbose=config['verbose'])
+
 
     # Setup simmilarity matrix
     adjmat = bnlearn._dag2adjmat(out['model'])
@@ -218,18 +242,28 @@ def fit(df, methodtype='hc', scoretype='bic', black_list=None, white_list=None,
 
 
 # %% white_list and black_list
-def _white_black_list_filter(df, white_list, black_list, bw_list_method='enforce', verbose=3):
-    if bw_list_method=='filter':
+def _white_black_list_filter(df, white_list, black_list, bw_list_method='edges', verbose=3):
+    # if bw_list_method=='edges':
+    #     # Keep only edges that are in white_list.
+    #     if white_list is not None:
+    #         if verbose>=3: print('[bnlearn] >Filter variables on white_list..')
+    #         parent = [ u  for (u, v) in white_list]
+    #         child = [ v  for (u, v) in white_list]
+    #         white_list_node = [x.lower() for x in set(parent+child)]
+    #         Iloc = np.isin(df.columns.str.lower(), white_list_node)
+    #         df = df.loc[:, Iloc]
+
+    if bw_list_method=='nodes':
         # Keep only variables that are in white_list.
         if white_list is not None:
-            if verbose>=3: print('[bnlearn] >Filter variables on white_list..')
+            if verbose>=3: print('[bnlearn] >Filter variables (nodes) on white_list..')
             white_list = [x.lower() for x in white_list]
             Iloc = np.isin(df.columns.str.lower(), white_list)
             df = df.loc[:, Iloc]
 
         # Exclude variables that are in black_list.
         if black_list is not None:
-            if verbose>=3: print('[bnlearn] >Filter variables on black_list..')
+            if verbose>=3: print('[bnlearn] >Filter variables (nodes) on black_list..')
             black_list = [x.lower() for x in black_list]
             Iloc = ~np.isin(df.columns.str.lower(), black_list)
             df = df.loc[:, Iloc]
@@ -240,19 +274,19 @@ def _white_black_list_filter(df, white_list, black_list, bw_list_method='enforce
     return df
 
 
-# %% hillclimbsearch
-def _chowliu(df, root_node, verbose=3):
-    """Chow-liu searches for DAGs with attempts to find a model with optimal score.
+# %% TreeSearch methods
+def _treesearch(df, estimator_type, root_node, class_node=None, verbose=3):
+    """Tree search methods.
 
     Description
     -----------
+    The TreeSearch methods Chow-liu and TAN (Tree-augmented Naive Bayes)
+    searches for DAGs with attempts to find a model with optimal score.
 
     """
-    from pgmpy.estimators import TreeSearch
-
     out={}
     est = TreeSearch(df, root_node=root_node)
-    model = est.estimate(estimator_type="chow-liu")
+    model = est.estimate(estimator_type=estimator_type, class_node=class_node)
 
     # Store
     out['model']=model
@@ -265,12 +299,28 @@ def _chowliu(df, root_node, verbose=3):
 def _constraintsearch(df, significance_level=0.05, verbose=3):
     """Contrain search.
 
+    PC PDAG construction is only guaranteed to work under the assumption that the
+    identified set of independencies is *faithful*, i.e. there exists a DAG that
+    exactly corresponds to it. Spurious dependencies in the data set can cause
+    the reported independencies to violate faithfulness. It can happen that the
+    estimated PDAG does not have any faithful completions (i.e. edge orientations
+    that do not introduce new v-structures). In that case a warning is issued.
+
     test_conditional_independence() returns a tripel (chi2, p_value, sufficient_data),
     consisting in the computed chi2 test statistic, the p_value of the test, and a heuristig
     flag that indicates if the sample size was sufficient.
     The p_value is the probability of observing the computed chi2 statistic (or an even higher chi2 value),
     given the null hypothesis that X and Y are independent given Zs.
     This can be used to make independence judgements, at a given level of significance.
+
+    DAG (pattern) construction
+    With a method for independence testing at hand, we can construct a DAG from the data set in three steps:
+        1. Construct an undirected skeleton - `estimate_skeleton()`
+        2. Orient compelled edges to obtain partially directed acyclid graph (PDAG; I-equivalence class of DAGs) - `skeleton_to_pdag()`
+        3. Extend DAG pattern to a DAG by conservatively orienting the remaining edges in some way - `pdag_to_dag()`
+
+        Step 1.&2. form the so-called PC algorithm, see [2], page 550. PDAGs are `DirectedGraph`s, that may contain both-way edges, to indicate that the orientation for the edge is not determined.
+
     """
     out = {}
     # Set search algorithm
@@ -281,23 +331,14 @@ def _constraintsearch(df, significance_level=0.05, verbose=3):
     #    print(_is_independent(est, 'Cloudy', 'Rain', significance_level=significance_level))
     #    print(_is_independent(est, 'Sprinkler', 'Rain',  ['Wet_Grass'], significance_level=significance_level))
 
-    """
-    DAG (pattern) construction
-    With a method for independence testing at hand, we can construct a DAG from the data set in three steps:
-        1. Construct an undirected skeleton - `estimate_skeleton()`
-        2. Orient compelled edges to obtain partially directed acyclid graph (PDAG; I-equivalence class of DAGs) - `skeleton_to_pdag()`
-        3. Extend DAG pattern to a DAG by conservatively orienting the remaining edges in some way - `pdag_to_dag()`
-
-        Step 1.&2. form the so-called PC algorithm, see [2], page 550. PDAGs are `DirectedGraph`s, that may contain both-way edges, to indicate that the orientation for the edge is not determined.
-    """
     # Estimate using chi2
     [skel, seperating_sets] = model.build_skeleton(significance_level=significance_level)
 
-    print("Undirected edges: ", skel.edges())
+    if verbose>=4: print("Undirected edges: ", skel.edges())
     pdag = model.skeleton_to_pdag(skel, seperating_sets)
-    print("PDAG edges: ", pdag.edges())
+    if verbose>=4: print("PDAG edges: ", pdag.edges())
     dag = pdag.to_dag()
-    print("DAG edges: ", dag.edges())
+    if verbose>=4: print("DAG edges: ", dag.edges())
 
     out['undirected'] = skel
     out['undirected_edges'] = skel.edges()
@@ -311,21 +352,12 @@ def _constraintsearch(df, significance_level=0.05, verbose=3):
     out['model'] = best_model
     out['model_edges'] = best_model.edges()
 
-    print(best_model.edges())
-
-    """
-    PC PDAG construction is only guaranteed to work under the assumption that the
-    identified set of independencies is *faithful*, i.e. there exists a DAG that
-    exactly corresponds to it. Spurious dependencies in the data set can cause
-    the reported independencies to violate faithfulness. It can happen that the
-    estimated PDAG does not have any faithful completions (i.e. edge orientations
-    that do not introduce new v-structures). In that case a warning is issued.
-    """
+    if verbose>=4: print(best_model.edges())
     return(out)
 
 
 # %% hillclimbsearch
-def _hillclimbsearch(df, scoretype='bic', black_list=None, white_list=None, max_indegree=None, tabu_length=100, epsilon=1e-4, max_iter=1e6, bw_list_method='enforce', verbose=3):
+def _hillclimbsearch(df, scoretype='bic', black_list=None, white_list=None, max_indegree=None, tabu_length=100, epsilon=1e-4, max_iter=1e6, bw_list_method='edges', fixed_edges=set(), verbose=3):
     """Heuristic hill climb searches for DAGs, to learn network structure from data. `estimate` attempts to find a model with optimal score.
 
     Description
@@ -355,14 +387,14 @@ def _hillclimbsearch(df, scoretype='bic', black_list=None, white_list=None, max_
     model = HillClimbSearch(df)
 
     # Compute best DAG
-    if bw_list_method=='enforce':
+    if bw_list_method=='edges':
         if (black_list is not None) or (white_list is not None):
-            if verbose>=3: print('[bnlearn] >Enforcing nodes based on black_list and/or white_list.')
+            if verbose>=3: print('[bnlearn] >Filter edges based on black_list/white_list')
         # best_model = model.estimate()
-        best_model = model.estimate(scoring_method=scoring_method, max_indegree=max_indegree, tabu_length=tabu_length, epsilon=epsilon, max_iter=max_iter, black_list=black_list, white_list=white_list, show_progress=False)
+        best_model = model.estimate(scoring_method=scoring_method, max_indegree=max_indegree, tabu_length=tabu_length, epsilon=epsilon, max_iter=max_iter, black_list=black_list, white_list=white_list, fixed_edges=fixed_edges, show_progress=False)
     else:
         # At this point, variables are readily filtered based on bw_list_method or not (if nothing defined).
-        best_model = model.estimate(scoring_method=scoring_method, max_indegree=max_indegree, tabu_length=tabu_length, epsilon=epsilon, max_iter=max_iter, show_progress=False)
+        best_model = model.estimate(scoring_method=scoring_method, max_indegree=max_indegree, tabu_length=tabu_length, epsilon=epsilon, max_iter=max_iter, fixed_edges=fixed_edges, show_progress=False)
 
     # Store
     out['model']=best_model
